
//---Author:  Richard T. Jones  07/12/2007
//---Version:  1.0

/*************************************************************************
 * Copyright(c) 1998, University of Connecticut, All rights reserved.    *
 * Author: Richard T. Jones, Asst. Prof. of Physics                      *
 *                                                                       *
 * Permission to use, copy, modify and distribute this software and its  *
 * documentation for non-commercial purposes is hereby granted without   *
 * fee, provided that the above copyright notice appears in all copies   *
 * and that both the copyright notice and this permission notice appear  *
 * in the supporting documentation. The author makes no claims about the *
 * suitability of this software for any purpose.                         *
 * It is provided "as is" without express or implied warranty.           *
 *************************************************************************/
//////////////////////////////////////////////////////////////////////////
//									//
// Michelson Interferometric Topography Analysis Package		//
//									//
// The present package implements tools for modeling the shape of a     //
// general 2D surface in terms of its Legendre expansion.		//
//									//
//////////////////////////////////////////////////////////////////////////
 
#include <iostream>

using namespace std;

#include <TASImage.h>
#include <TArrayD.h>
#include <Math/SpecFuncMathMore.h>
#include <TBitmap.h>

ClassImp(TBitmap)


TBitmap& TBitmap::operator+=(const TBitmap &source)
{
   if (source.fWidth != fWidth || source.fHeight != fHeight) {
      std::cerr << "Error in TBitmap::operator+= : source and destination maps"
                << " have different dimensions!" << std::endl;
      return *this;
   }

   UInt_t k=0;
   for (UInt_t m=0; m<fHeight; m++) {
      for (UInt_t n=0; n<fWidth; n++,k++) {
         fImage[k] += source.fImage[k];
      }
   }
   if (fDeg)
      Precompute(fDeg);
   if (fTimage) {
      delete fTimage;
      fTimage = 0;
   }
   return *this;
}

TBitmap& TBitmap::operator-=(const TBitmap &source)
{
   if (source.fWidth != fWidth || source.fHeight != fHeight) {
      std::cerr << "Error in TBitmap::operator-= : source and destination maps"
                << " have different dimensions!" << std::endl;
      return *this;
   }

   UInt_t k=0;
   for (UInt_t m=0; m<fHeight; m++) {
      for (UInt_t n=0; n<fWidth; n++,k++) {
         fImage[k] -= source.fImage[k];
      }
   }
   if (fDeg)
      Precompute(fDeg);
   if (fTimage) {
      delete fTimage;
      fTimage = 0;
   }
   return *this;
}

TBitmap& TBitmap::operator*=(const TBitmap &source)
{
   if (source.fWidth != fWidth || source.fHeight != fHeight) {
      std::cerr << "Error in TBitmap::operator*= : source and destination maps"
                << " have different dimensions!" << std::endl;
      return *this;
   }

   UInt_t k=0;
   for (UInt_t m=0; m<fHeight; m++) {
      for (UInt_t n=0; n<fWidth; n++,k++) {
         fImage[k] *= source.fImage[k];
      }
   }
   if (fDeg)
      Precompute(fDeg);
   if (fTimage) {
      delete fTimage;
      fTimage = 0;
   }
   return *this;
}

TBitmap& TBitmap::operator/=(const TBitmap &source)
{
   if (source.fWidth != fWidth || source.fHeight != fHeight) {
      std::cerr << "Error in TBitmap::operator/= : source and destination maps"
                << " have different dimensions!" << std::endl;
      return *this;
   }

   UInt_t k=0;
   for (UInt_t m=0; m<fHeight; m++) {
      for (UInt_t n=0; n<fWidth; n++,k++) {
         fImage[k] /= (source.fImage[k] + 1e-30);
      }
   }
   if (fDeg)
      Precompute(fDeg);
   if (fTimage) {
      delete fTimage;
      fTimage = 0;
   }
   return *this;
}

TBitmap& TBitmap::operator*=(Double_t factor)
{
   UInt_t k=0;
   for (UInt_t m=0; m<fHeight; m++) {
      for (UInt_t n=0; n<fWidth; n++,k++) {
         fImage[k] *= factor;
      }
   }
   if (fDeg)
      Precompute(fDeg);
   if (fTimage) {
      delete fTimage;
      fTimage = 0;
   }
   return *this;
}

TBitmap& TBitmap::operator/=(Double_t factor)
{
   UInt_t k=0;
   for (UInt_t m=0; m<fHeight; m++) {
      for (UInt_t n=0; n<fWidth; n++,k++) {
         fImage[k] /= (factor+1e-30);
      }
   }
   if (fDeg)
      Precompute(fDeg);
   if (fTimage) {
      delete fTimage;
      fTimage = 0;
   }
   return *this;
}

TBitmap TBitmap::operator-() const
{
   TBitmap map(*this);
   UInt_t k=0;
   for (UInt_t m=0; m<fHeight; m++) {
      for (UInt_t n=0; n<fWidth; n++,k++) {
         fImage[k] = -fImage[k];
      }
   }
   if (fDeg)
      map.Precompute(fDeg);
   return map;
}

Bool_t TBitmap::operator==(const TBitmap &other) const
{
   if (other.fWidth != fWidth || other.fHeight != fHeight) {
      return kFALSE;
   }

   UInt_t k=0;
   for (UInt_t m=0; m<fHeight; m++) {
      for (UInt_t n=0; n<fWidth; n++,k++) {
         if (fImage[k] != other.fImage[k]) {
            return kFALSE;
         }
      }
   }
   return kTRUE;
}

void TBitmap::Compute(const TModel& model)
{
   UInt_t k=0;
   for (UInt_t m=0; m<fHeight; m++) {
      Double_t y=(m-fHeight/2.)/fHeight;
      for (UInt_t n=0; n<fWidth; n++,k++) {
         Double_t x=(n-fWidth/2.)/fWidth;
         fImage[k] = model(x,y);
      }
   }
   if (fTimage) {
      delete fTimage;
      fTimage = 0;
   }
}

void TBitmap::Precompute(UInt_t degree)
{
   // Precompute the Legendre basis functions at the center of each pixel

   if (fDeg) {
      delete [] fXbasis;
      delete [] fYbasis;
   }
   if (degree == 0) {
      fDeg = 0;
      return;
   }
   fXbasis = new Double_t[fWidth*degree];
   fYbasis = new Double_t[fHeight*degree];
   fDeg = degree;
   UInt_t k=0;
   for (UInt_t n=0; n<fWidth; n++) {
      Double_t x=(n-fWidth/2.)/fWidth;
      for (UInt_t m=0; m<degree; m++,k++) {
         fXbasis[k] = ROOT::Math::legendre(m,x);
      }
   }
   k=0;
   for (UInt_t n=0; n<fHeight; n++) {
      Double_t y=(n-fHeight/2.)/fHeight;
      for (UInt_t m=0; m<degree; m++,k++) {
         fYbasis[k] = ROOT::Math::legendre(m,y);
      }
   }
}

void TBitmap::Draw(Option_t* option)
{
   if (fTimage == 0) {
      fTimage = new TASImage(fWidth,fHeight);
      fTimage->SetImage(fImage,fWidth,fHeight);
   }
   // fTimage->Gray();
   fTimage->Draw(option);
}

void TBitmap::InterfereFast(const TModel& refa,
                            const TModel& sig1a,
                            const TModel& sig2a,
                            const TModel& sig1ph,
                            const TModel& sig2ph)
{
   // A faster version of Interfere()
   // Assumes Precompute(deg) is called first with a degree deg at least
   // as large as the highest-order model in the argument list, otherwise
   // the following implementation truncates all models at degree deg.

   // std::cout << "+" << std::endl;
   if (fTimage) {
      delete fTimage;
      fTimage = 0;
   }
   UInt_t drefa=refa.GetDegree();
   UInt_t dsig1a=sig1a.GetDegree();
   UInt_t dsig2a=sig2a.GetDegree();
   UInt_t dsig1ph=sig1ph.GetDegree();
   UInt_t dsig2ph=sig2ph.GetDegree();
   UInt_t km=0;
   for (UInt_t m=0; m<fHeight; m++, km+=fDeg) {
      UInt_t kn=0;
      for (UInt_t n=0; n<fWidth; n++, kn+=fDeg) {
         Double_t rrefa=0;
         Double_t rsig1a=0;
         Double_t rsig2a=0;
         Double_t rsig1ph=0;
         Double_t rsig2ph=0;
         UInt_t ki=km;
         UInt_t kj=kn;
         for (UInt_t i=0; i<fDeg; i++,ki++,kj=kn) {
            for (UInt_t j=0; j<fDeg; j++,kj++) {
               Double_t basis = fXbasis[kj]*fYbasis[ki];
               if (i < drefa && j < drefa)
                  rrefa += refa[j][i]*basis;
               if (i < dsig1a && j < dsig1a)
                  rsig1a += sig1a[j][i]*basis;
               if (i < dsig2a && j < dsig2a)
                  rsig2a += sig2a[j][i]*basis;
               if (i < dsig1ph && j < dsig1ph)
                  rsig1ph += sig1ph[j][i]*basis;
               if (i < dsig2ph && j < dsig2ph)
                  rsig2ph += sig2ph[j][i]*basis;
            }
         }
         fImage[m*fWidth+n] = rrefa*rrefa
                            + rsig1a*rsig1a
                            + rsig2a*rsig2a
                            + 2*rrefa*rsig1a*cos(rsig1ph)
                            + 2*rrefa*rsig2a*cos(rsig2ph)
                            + 2*rsig1a*rsig2a*cos(rsig1ph-rsig2ph);
      }
   }
   //std::cout << "=" << std::endl;
}

void TBitmap::Interfere(const TModel& refa,
                        const TModel& sig1a,
                        const TModel& sig2a,
                        const TModel& sig1ph,
                        const TModel& sig2ph)
{
   // Form a Michelson Interferometer image using
   //     refa: amplitude map of reference beam
   //     sig1a: amplitude map of signal beam 1
   //     sig2a: amplitude map of signal beam 2
   //     sig1ph: phase map of signal beam 1 (relative to ref beam)
   //     sig2ph: phase map of signal beam 2 (relative to ref beam)

   if (fDeg) {
      InterfereFast(refa,sig1a,sig2a,sig1ph,sig2ph);
      return;
   }

   std::cout << "++" << std::endl;
   if (fTimage) {
      delete fTimage;
      fTimage = 0;
   }
   for (UInt_t m=0; m<fHeight; m++) {
      Double_t y=(m-fHeight/2.)/fHeight;
      for (UInt_t n=0; n<fWidth; n++) {
         Double_t x=(n-fWidth/2.)/fWidth;
         Double_t drefa=refa(x,y);
         Double_t dsig1a=sig1a(x,y);
         Double_t dsig2a=sig2a(x,y);
         Double_t dsig1ph=sig1ph(x,y);
         Double_t dsig2ph=sig2ph(x,y);
         fImage[m*fWidth+n] = drefa*drefa
                            + dsig1a*dsig1a
                            + dsig2a*dsig2a
                            + 2*drefa*dsig1a*cos(dsig1ph)
                            + 2*drefa*dsig2a*cos(dsig2ph)
                            + 2*dsig1a*dsig2a*cos(dsig1ph-dsig2ph);
      }
   }
   std::cout << "==" << std::endl;
}

Double_t TBitmap::ChiSquared(const TBitmap& map) const
{
   // Form a chi-square comparison between this image and the map argument

   if (map.fWidth != fWidth || map.fHeight != fHeight) {
      std::cerr << "Error in TBitmap::ChiSquared : this and map images"
                << " have different dimensions!" << std::endl;
      return 0;
   }
   Double_t chi2 = 0;
   for (UInt_t n=0; n<fWidth*fHeight; n++) {
      Double_t diff = fImage[n]-map.fImage[n];
      chi2 += diff*diff;
   }
   return chi2;
}

void TBitmap::Streamer(TBuffer &buf)
{
   // Put/get image data to/from stream buffer buf.

   if (buf.IsReading()) {
      buf >> fWidth >> fHeight;
      buf.ReadStaticArray(fImage);
   } else {
      buf << fWidth << fHeight;
      buf.WriteArray(fImage, fWidth*fHeight);
   }
}

TBitmap operator+(const TBitmap& v1, const TBitmap& v2)
{
   TBitmap map(v1);
   return map+=v2;
}

TBitmap operator-(const TBitmap& v1, const TBitmap& v2)
{
   TBitmap map(v1);
   return map-=v2;
}

TBitmap operator*(const TBitmap& v1, const TBitmap& v2)
{
   TBitmap map(v1);
   return map*=v2;
}

TBitmap operator/(const TBitmap& v1, const TBitmap& v2)
{
   TBitmap map(v1);
   return map/=v2;
}

TBitmap operator*(const TBitmap& v1, Double_t factor)
{
   TBitmap map(v1);
   return map*=factor;
}

TBitmap operator*(Double_t factor, const TBitmap& v1)
{
   TBitmap map(v1);
   return map*=factor;
}

TBitmap operator/(const TBitmap& v1, Double_t factor)
{
   TBitmap map(v1);
   return map/=factor;
}
