//******************************************************
//  Copyright 1999 University of Paderborn
//
//  FILE : SA_MIProblem.cc
//  
//  See the file SA_MIProblem.h for documentation on this class.
//
//  author:
//    Richard Jones
//  contact:
//    jonesrt@phys.uconn.edu
//  last update:
//    July 25, 2007
//*********************************************************

#include "SA_MIProblem.h"

#include <fstream>
#include <iomanip>
#include <string>
#include <sstream>

SA_Solution::SA_Solution()
 : fRefa(0), fSig1a(0), fSig1ph(0), fSig2ph(0),
   fNdim(0), fCost(-1), fNewcost(-1), fLastmove(0)
{ }

SA_Solution::SA_Solution(const unsigned int ndim)
{
   fRefa = new TModel(ndim);
   fSig1a = new TModel(ndim);
   fSig1ph = new TModel(ndim);
   fSig2ph = new TModel(ndim);
   fNdim = ndim;
   fCost = -1;
   fLastmove = 0;
}

SA_Solution::~SA_Solution()
{
   if (fRefa)
      delete fRefa;
   if (fSig1a)
      delete fSig1a;
   if (fSig1ph)
      delete fSig1ph;
   if (fSig2ph)
      delete fSig2ph;
   if (fLastmove)
      delete fLastmove;
} 

void SA_Solution::Copy(SA_Solution& source) 
{ 
   // Implements deep-copy sematics for SA_Solution

   if (fRefa) {
      delete fRefa;
      fRefa = 0;
   }
   if (fSig1a) {
      delete fSig1a;
      fSig1a = 0;
   }
   if (fSig1ph) {
      delete fSig1ph;
      fSig1ph = 0;
   }
   if (fSig2ph) {
      delete fSig2ph;
      fSig2ph = 0;
   }
   if (fLastmove) {
      delete fLastmove;
      fLastmove = 0;
   }
   fNdim = source.fNdim;
   if (fNdim) {
      fRefa = new TModel(*(source.fRefa));
      fSig1a = new TModel(*(source.fSig1a));
      fSig1ph = new TModel(*(source.fSig1ph));
      fSig2ph = new TModel(*(source.fSig2ph));
   }
   fCost = source.fCost;
   fNewcost = source.fNewcost;
   if (source.fLastmove) {
      fLastmove = new SA_Move(*(source.fLastmove));
   }
}

std::ostream& operator<<(std::ostream& out, SA_Solution& sol) 
{
   out << " " << sol.fNdim;
   out << " " << sol.fCost;
   out << " " << sol.fNewcost;
   out << sol.fRefa;
   out << sol.fSig1a;
   out << sol.fSig1ph;
   out << sol.fSig2ph;
   if (sol.fLastmove) {
      out << " 1" << *(sol.fLastmove);
   }
   else {
      out << " 0";
   }
   return out; 
}

std::istream& operator>>(std::istream& in, SA_Solution*& sol) 
{
   sol = new SA_Solution();
   in >> sol->fNdim >> sol->fCost >> sol->fNewcost;
   in >> sol->fRefa;
   in >> sol->fSig1a;
   in >> sol->fSig1ph;
   in >> sol->fSig2ph;
   int has_lastmove;
   in >> has_lastmove;
   if (has_lastmove) {
      in >> sol->fLastmove;
   }
   else {
      sol->fLastmove = 0;
   }
   return(in); 
}

SA_Move::SA_Move()
 : fModel_set(0), fModel(0) {}

SA_Move::~SA_Move()
{
   if (fModel)
      delete fModel;
}

SA_Move::SA_Move(SA_Problem& prob)
 : fModel_set(0)
{
   int ndim = prob.GetNdim();
   fModel = new TModel(ndim);
}

SA_Move::SA_Move(const SA_Move& move)
{
   fModel_set = move.fModel_set;
   if (move.fModel) {
      fModel = new TModel(*(move.fModel));
   }
   else {
      fModel = 0;
   }
}

std::ostream& operator<<(std::ostream& out, SA_Move& move) 
{
   out << " " << move.fModel_set;
   if (move.fModel) {
      out << " 1" << move.fModel;
   }
   else {
      out << " 0";
   }
   return out; 
}

std::istream& operator>>(std::istream& in, SA_Move*& move) 
{
   move = new SA_Move();
   in >> move->fModel_set;
   int has_fModel;
   in >> has_fModel;
   if (has_fModel) {
      in >> move->fModel;
   }
   else {
      move->fModel = 0;
   }
   return(in); 
}

SA_Problem::SA_Problem(int ndim) 
 : fDataImage(0), fModelImage(0), fNdim(ndim)
{ 
   for (int i=0; i<4; i++) {
      fNvar[i] = ndim;
   }
}

SA_Problem::~SA_Problem()
{
   if (fDataImage)
      delete fDataImage;
   if (fModelImage)
      delete fModelImage;
}

void SA_Problem::EvaluateOptions(int *argc,char ***argv)
{
   // Picks through the command line for options relevant to SA_Problem

   for (int j=0; j<*argc; j++) {
      if (strstr((*argv)[j],"--ndim") == (*argv)[j]) {
         fNdim = atoi((*argv)[++j]);
      }
      else if (strstr((*argv)[j],"--nvar") == (*argv)[j]) {
         std::istringstream str((*argv)[++j]);
         for (int i=0; i<4; i++) {
            char s[100];
            str.getline(s,99,',');
            fNvar[i] = atoi(s);
         }
      }
      else if (strstr((*argv)[j],"--init") == (*argv)[j]) {
         std::istringstream str((*argv)[++j]);
         str >> fInitSol;
      }
   }
}

int SA_Problem::ReadProblemData(char *filename)
{   
   // Reads the problem data from file "filename".

   if (fDataImage) {
      delete fDataImage;
      fDataImage = 0;
   }
   if (fModelImage) {
      delete fModelImage;
      fModelImage = 0;
   }
   std::ifstream initfile(filename);
   if (initfile.good()) {
      initfile >> fDataImage;
      fModelImage = new TBitmap(fDataImage->GetWidth(),fDataImage->GetHeight());
      fModelImage->Precompute(fNdim);
      return 1;
   }
   return 0;
}
	
SA_Solution * SA_Problem::CreateSolution()
{ 
   // Create a new SA_Solution instance for this problem;
   // the user is responsible for deleting it when it is no longer needed.

   SA_Solution* newsol = new SA_Solution(fNdim);
   return newsol;
}

void SA_Problem::GetInitialSolution(SA_Solution& sol)
{ 
  // Set the SA_Solution sol to a predefined initialization value

   std::ifstream in(fInitSol.c_str());
   if (in.good()) {
      SA_Solution* s;
      in >> s;
      Copy(*s,sol);
      delete s;
      return;
   }

   if (sol.fNdim > 0) {
      UInt_t drefa = sol.fRefa->GetDegree();
      UInt_t dsig1a = sol.fSig1a->GetDegree();
      UInt_t dsig1ph = sol.fSig1ph->GetDegree();
      UInt_t dsig2ph = sol.fSig2ph->GetDegree();
      for (int i=0; i<drefa; i++)
         for (int j=0; j<drefa; j++)
            (*sol.fRefa)[i][j] = 0;
      for (int i=0; i<dsig1a; i++)
         for (int j=0; j<dsig1a; j++)
            (*sol.fSig1a)[i][j] = 0;
      for (int i=0; i<dsig1ph; i++)
         for (int j=0; j<dsig1ph; j++)
            (*sol.fSig1ph)[i][j] = 0;
      for (int i=0; i<dsig2ph; i++)
         for (int j=0; j<dsig2ph; j++)
            (*sol.fSig2ph)[i][j] = 0;
      (*sol.fRefa)[0][0] = 1;
      (*sol.fSig1a)[0][0] = 1;
      sol.fCost = ComputeCost(sol);
      sol.fNewcost = -1;
      if (sol.fLastmove) {
         delete sol.fLastmove;
         sol.fLastmove = 0;
      }
   }
}

void SA_Problem::GetRandomSolution(SA_Solution& sol)
{
  // Set the SA_Solution sol to a random value

   if (sol.fNdim > 0) {
      sol.fRefa->Randomize(fRefa_scale);
      sol.fSig1a->Randomize(fSig1a_scale);
      sol.fSig1ph->Randomize(fSig1ph_scale);
      sol.fSig2ph->Randomize(fSig2ph_scale);
      sol.fCost = ComputeCost(sol);
      sol.fNewcost = -1;
      if (sol.fLastmove) {
         delete sol.fLastmove;
         sol.fLastmove = 0;
      }
   }
}

double SA_Problem::ComputeCost(SA_Solution& sol)
{
   // Compute the cost function for this solution

   double cost = -1;
   if (fModelImage && (sol.fNdim == fNdim)) {
      TModel* sig2a = new TModel(*(sol.fSig1a));
      *sig2a *= fSig2a_ratio;
      fModelImage->Interfere(*(sol.fRefa),*(sol.fSig1a),*sig2a,
                             *(sol.fSig1ph),*(sol.fSig2ph));
      cost = fModelImage->ChiSquared(*fDataImage);
      delete sig2a;
   }
   return cost;
}

float SA_Problem::GetCost(SA_Solution& sol)
{
   // Return the pre-computed cost for this solution

   if (sol.fLastmove)
      return sol.fNewcost;
   else
      return sol.fCost;
}

void SA_Problem::GetNeighbor(SA_Solution& sol, float rel_temp)
{
   // Generate a reversible random change in the given 
   // solution. The range of change can be modified 
   // depending on relative temperature rel_temp
   // which ranges from 0 (=low) to 1 (=high).

   SA_Move* move = new SA_Move();
   do {
      long randomnumber;
      randomnumber = rand();
      move->fModel_set = randomnumber % 4;
   } while (fNvar[move->fModel_set] == 0);

   TModel* orig;
   double rescale = pow(rel_temp,0.3);
   switch (move->fModel_set) {
      case kRefa:
         orig = new TModel(*(sol.fRefa));
         sol.fRefa->Mutate(rescale*fRefa_step,fNvar[kRefa]);
         sol.fNewcost = ComputeCost(sol);
         move->fModel = sol.fRefa;
         sol.fRefa = orig;
         break;
      case kSig1a:
         orig = new TModel(*(sol.fSig1a));
         sol.fSig1a->Mutate(rescale*fSig1a_step,fNvar[kSig1a]);
         sol.fNewcost = ComputeCost(sol);
         move->fModel = sol.fSig1a;
         sol.fSig1a = orig;
         break;
      case kSig1ph:
         orig = new TModel(*(sol.fSig1ph));
         sol.fSig1ph->Mutate(rescale*fSig1ph_step,fNvar[kSig1ph]);
         sol.fNewcost = ComputeCost(sol);
         move->fModel = sol.fSig1ph;
         sol.fSig1ph = orig;
         break;
      case kSig2ph:
         orig = new TModel(*(sol.fSig2ph));
         sol.fSig2ph->Mutate(rescale*fSig2ph_step,fNvar[kSig2ph]);
         sol.fNewcost = ComputeCost(sol);
         move->fModel = sol.fSig2ph;
         sol.fSig2ph = orig;
         break;
      default:
         std::cerr << "Fatal error in SA_Problem::GetNeighbor, "
                   << "this should never happen!" << std::endl;
         exit(1);
   }
   if (sol.fLastmove)
      delete sol.fLastmove;
   sol.fLastmove = move;
}

void SA_Problem::ResetSolution(SA_Solution& sol)
{
   // Discards the last change of solution made by GetNeighbor()

   if (sol.fLastmove) {
      delete sol.fLastmove;
      sol.fLastmove = 0;
   }
   sol.fNewcost = -1;
}

void SA_Problem::UpdateSolution(SA_Solution& sol)
{
   // Make the last move made by GetNeighbor() permanent

   if (sol.fLastmove == 0)
      return;

   switch (sol.fLastmove->fModel_set) {
      case kRefa:
         delete sol.fRefa;
         sol.fRefa = sol.fLastmove->fModel;
         break;
      case kSig1a:
         delete sol.fSig1a;
         sol.fSig1a = sol.fLastmove->fModel;
         break;
      case kSig1ph:
         delete sol.fSig1ph;
         sol.fSig1ph = sol.fLastmove->fModel;
         break;
      case kSig2ph:
         delete sol.fSig2ph;
         sol.fSig2ph = sol.fLastmove->fModel;
         break;
      default:
         std::cerr << "Fatal error in SA_Problem::GetNeighbor, "
                   << "this should never happen!" << std::endl;
   }
   sol.fCost = sol.fNewcost;
   sol.fNewcost = -1;
   sol.fLastmove = 0;
}

float SA_Problem::GetLocalN()
{ 
   // Return  the number of solution neighbours (approximately).
   // The scaling with problem size is correct, but the magnitude
   // has been arbitrarily fixed to fit within the float range
   // for most problems.

   double nearest_neighbors = fNvar[0]*fNvar[0]
                            + fNvar[1]*fNvar[1]
                            + fNvar[2]*fNvar[2]
                            + fNvar[3]*fNvar[3];
   return 2*nearest_neighbors;
}

float SA_Problem::GetGlobalN()
{ 
   // Return the total number of distinguishable solutions (approximately).
   // The scaling with problem size is correct, but the magnitude
   // has been arbitrarily fixed to fit within the float range
   // for most problems.

   double logN = log(fRefa_scale/fRefa_step)*fNvar[0]*fNvar[0]
               + log(fSig1a_scale/fSig1a_step)*fNvar[1]*fNvar[1]
               + log(fSig1ph_scale/fSig1ph_step)*fNvar[2]*fNvar[2]
               + log(fSig2ph_scale/fSig2ph_step)*fNvar[3]*fNvar[3];
   return (logN < log(FLT_MAX))? exp(logN) : FLT_MAX;
}

void SA_Problem::Copy(SA_Solution& source, SA_Solution& dest)
{
   // Implements deep-copy semantics for SA_Solution

   dest.Copy(source);
}	

void SA_Problem::OutputSolution(std::ostream& out, SA_Solution& sol)
{ 
   // Serialize a solution into an output stream

  out << sol;
}

int SA_Problem::InputSolution(std::istream& in, SA_Solution& sol)
{
   // Reads a solution from an input stream

   SA_Solution* s;
   in >> s;
   Copy(*s,sol);
   delete s;
   return (in.gcount() > 0);
}

#ifdef MPI

int  SA_Problem::BcastSolution(SA_Solution& sol, int root,
                               MPI_Comm comm, int me_comm)
{ 
   // Broadcast a solution to all MPI slave processors

   std::cerr << "SA_Problem::BcastSolution not yet implemented!" << std::endl;
   return 0;
}

#endif

SA_Move* SA_Problem::CreateMove()
{
   // Create a new instance of SA_Move for this problem;
   // it is up to the user to delete it when it is no longer needed.

   return new SA_Move();
}

int SA_Problem::ExtractMove(SA_Solution& sol, SA_Move& mov)
{
   // Extract the current move from sol and stores it in mov

   if (sol.fLastmove) {
      mov.fModel_set = sol.fLastmove->fModel_set;
      if (mov.fModel)
         delete mov.fModel;
      mov.fModel = new TModel(*(sol.fLastmove->fModel));
      return 1;
   }
   return 0;
}

int SA_Problem::ApplyMove(SA_Solution& sol, SA_Move& mov)	
{
   // applies the move mov to solution sol

   if (sol.fLastmove)
      delete sol.fLastmove;
   sol.fLastmove = new SA_Move(mov);
   UpdateSolution(sol);
   sol.fCost = ComputeCost(sol);
   return 1;
}

void SA_Problem::OutputMove(std::ostream& out, SA_Move& mov)
{
   // Write a move to an output stream

   out << mov;
}

int SA_Problem::InputApplyMoves(std::istream& in, SA_Solution& sol)
{
   // Read a number of moves from stream and change solution accordingly,
   // returning the number of steps successfully completed

   if (sol.fLastmove) {
      delete sol.fLastmove;
      sol.fLastmove = 0;
   }
   int nseen=0;
   while (!in.eof()) {
      in >> sol.fLastmove;
      UpdateSolution(sol);
      nseen++;
   }
   sol.fCost = ComputeCost(sol);
   return nseen;
}

#ifdef MPI

int SA_Problem::BcastApplyMove(SA_Solution& sol, int root,
                               MPI_Comm comm, int me_comm)
{
   // Broadcast a move and apply it to all slave processors

   std::cerr << "SA_Problem::BcastApplyMove not yet implemented!" << std::endl;
   return 0;
}

#endif

int SA_Problem::OutputSolution(char* filename, SA_Solution& sol)
{
   // Writes a solution to a file

   std::ofstream outfile(filename);
   if (!outfile) {
      std::cerr << "Error opening file: " << filename << std::endl;
      return 0;
   }
   else {
      outfile << sol;
      outfile.close();
   }
   return 1;
}

int SA_Problem::InputSolution(char* filename, SA_Solution& sol)
{
   // Reads a solution from a file

   std::ifstream infile(filename);
   if (!infile) {
      std::cerr << "Error opening file: " << filename << std::endl;
      return 0;
   }
   else {
      SA_Solution* s;
      infile >> s;
      Copy(*s,sol);
      delete s;
      infile.close();
   }
   return 1;
}

void SA_Problem::Postprocessing(SA_Solution& sol)
{
}

int SA_Problem::Reoptimize(SA_Solution& sol)
{
   return 1;
}

void SA_Problem::ControlOutput(std::ostream& out)
{
}
